# validated: 2018-11-10 EN 0e9172f9a708 edu/wpi/first/wpilibj/TimedRobot.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from typing import Optional

import hal

from .iterativerobotbase import IterativeRobotBase
from .notifier import Notifier
from .resource import Resource
from .robotcontroller import RobotController
from .timer import Timer


__all__ = ["TimedRobot"]


class TimedRobot(IterativeRobotBase):
    """TimedRobot implements the IterativeRobotBase robot program framework.

    The TimedRobot class is intended to be subclassed by a user creating a robot program.

    periodic() functions from the base class are called on an interval by a Notifier instance.
    """

    kDefaultPeriod = 0.02

    def __init__(self, period: Optional[float] = None) -> None:
        if period is None:
            period = TimedRobot.kDefaultPeriod
        super().__init__(period)
        hal.report(
            hal.UsageReporting.kResourceType_Framework,
            hal.UsageReporting.kFramework_Timed,
        )

        self._expirationTime = 0
        self._notifier = hal.initializeNotifier()

        Resource._add_global_resource(self)

    # python-specific
    def free(self) -> None:
        hal.stopNotifier(self._notifier)
        hal.cleanNotifier(self._notifier)

    def startCompetition(self) -> None:
        """Provide an alternate "main loop" via startCompetition()"""
        self.robotInit()

        hal.observeUserProgramStarting()

        self._expirationTime = RobotController.getFPGATime() * 1e-6 + self.period
        self._updateAlarm()

        # Loop forever, calling the appropriate mode-dependent function
        while True:
            if hal.waitForNotifierAlarm(self._notifier) == 0:
                break

            self._expirationTime += self.period
            self._updateAlarm()

            self.loopFunc()

    def getPeriod(self) -> float:
        """Get time period between calls to Periodic() functions."""
        return self.period

    def _updateAlarm(self) -> None:
        """Update the alarm hardware to reflect the next alarm."""
        hal.updateNotifierAlarm(self._notifier, int(self._expirationTime * 1e6))



# validated: 2018-10-30 EN 0a2ab4f0d782 edu/wpi/first/wpilibj/buttons/Trigger.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from ..command import command
from ..sendablebase import SendableBase
from ..sendablebuilder import SendableBuilder

__all__ = ["Trigger"]


class Trigger(SendableBase):
    """This class provides an easy way to link commands to inputs.

    It is very easy to link a button to a command.  For instance, you could
    link the trigger button of a joystick to a "score" command.

    It is encouraged that teams write a subclass of Trigger if they want to
    have something unusual (for instance, if they want to react to the user
    holding a button while the robot is reading a certain sensor input).
    For this, they only have to write the :func:`get` method to get the full
    functionality of the Trigger class.
    """

    def get(self) -> bool:
        """Returns whether or not the trigger is active

        This method will be called repeatedly a command is linked to the
        Trigger.

        :returns: whether or not the trigger condition is active.
        """
        raise NotImplementedError

    def grab(self) -> bool:
        """Returns whether :meth:`get` returns True or the internal table for
        :class:`.SmartDashboard` use is pressed.
        """
        return self.get() or getattr(self, "sendablePressed", False)

    def whenActive(self, command: "command.Command") -> None:
        """Starts the given command whenever the trigger just becomes active.

        :param command: the command to start
        """

        def execute() -> None:
            pressed = self.grab()

            if not execute.pressedLast and pressed:
                command.start()

            execute.pressedLast = pressed

        execute.pressedLast = self.grab()
        from ..command import Scheduler

        Scheduler.getInstance().addButton(execute)

    def whileActive(self, command: "command.Command") -> None:
        """Constantly starts the given command while the button is held.

        :meth:`Command.start` will be called repeatedly while the trigger is
        active, and will be canceled when the trigger becomes inactive.

        :param command: the command to start
        """

        def execute() -> None:
            pressed = self.grab()

            if pressed:
                command.start()
            elif execute.pressedLast and not pressed:
                command.cancel()

            execute.pressedLast = pressed

        execute.pressedLast = self.grab()
        from ..command import Scheduler

        Scheduler.getInstance().addButton(execute)

    def whenInactive(self, command: "command.Command") -> None:
        """Starts the command when the trigger becomes inactive.

        :param command: the command to start
        """

        def execute() -> None:
            pressed = self.grab()

            if execute.pressedLast and not pressed:
                command.start()

            execute.pressedLast = pressed

        execute.pressedLast = self.grab()
        from ..command import Scheduler

        Scheduler.getInstance().addButton(execute)

    def toggleWhenActive(self, command: "command.Command") -> None:
        """Toggles a command when the trigger becomes active.

        :param command: the command to toggle
        """

        def execute() -> None:
            pressed = self.grab()

            if not execute.pressedLast and pressed:
                if command.isRunning():
                    command.cancel()
                else:
                    command.start()

            execute.pressedLast = pressed

        execute.pressedLast = self.grab()
        from ..command import Scheduler

        Scheduler.getInstance().addButton(execute)

    def cancelWhenActive(self, command: "command.Command") -> None:
        """Cancels a command when the trigger becomes active.

        :param command: the command to cancel
        """

        def execute() -> None:
            pressed = self.grab()

            if not execute.pressedLast and pressed:
                command.cancel()

            execute.pressedLast = pressed

        execute.pressedLast = self.grab()
        from ..command import Scheduler

        Scheduler.getInstance().addButton(execute)

    def _safeState(self) -> None:
        self.sendablePressed = False

    def _setPressed(self, value: bool) -> None:
        self.sendablePressed = value

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("Button")
        builder.setSafeState(self._safeState)
        builder.addBooleanProperty("pressed", self.grab, self._setPressed)



# validated: 2018-11-18 EN 175c6c1f0130 edu/wpi/first/wpilibj/shuffleboard/ShuffleboardComponent.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from .container import ShuffleboardContainer
from typing import Dict, Any, Optional


class ShuffleboardComponent:
    def __init__(
        self, parent: ShuffleboardContainer, title: str, type: Optional[str] = None
    ):
        self.parent = parent
        self.title = title
        self.type = type
        self.metadataDirty = False
        self.column = -1
        self.row = -1
        self.width = -1
        self.height = -1
        self.properties = {}  # type: Dict[str, Any]

    def getParent(self) -> ShuffleboardContainer:
        return self.parent

    def setType(self, type: str) -> None:
        self.type = type
        self.metadataDirty = True

    def getType(self) -> str:
        return self.type

    def getTitle(self) -> str:
        return self.title

    def getProperties(self) -> Dict[str, Any]:
        return self.properties

    def withProperties(self, properties: Dict[str, Any]) -> "ShuffleboardComponent":
        """
        Sets custom properties for this component. Property names are case- 
        and whitespace-insensitive (capitalization and spaces do not matter).

        :param properties: the properties for this component

        :returns: this component
        """

        self.properties = properties
        self.metadataDirty = True
        return self

    def withPosition(self, columnIndex: int, rowIndex: int) -> "ShuffleboardComponent":
        """
        Sets the position of this component in the tab. This has no effect if 
        this component is inside a layout.

        If the position of a single component is set, it is recommended to set 
        the positions of *all* components inside a tab to prevent Shuffleboard 
        from automatically placing another component there before the one with 
        the specific position is sent.

        :param columnIndex: the column in the tab to place this component
        :param rowIndex:    the row in the tab to place this component
        :returns: this component
        """
        self.column = columnIndex
        self.row = rowIndex
        self.metadataDirty = True
        return self

    def withSize(self, width: int, height: int) -> "ShuffleboardComponent":
        """
        Sets the size of this component in the tab. This has no effect if this 
        component is inside a layout.

        :param width:  how many columns wide the component should be
        :param height: how many rows high the component should be
        :returns: this component
        """
        self.width = width
        self.height = height
        self.metadataDirty = True
        return self

    def buildMetadata(self, metaTable) -> None:
        if not self.metadataDirty:
            return

        # Component type
        if self.getType() is None:
            metaTable.getEntry("PreferredComponent").delete()
        else:
            metaTable.getEntry("PreferredComponent").forceSetString(self.getType())

        # Tile size
        if self.width <= 0 or self.height <= 0:
            metaTable.getEntry("Size").delete()
        else:
            metaTable.getEntry("Size").setNumberArray((self.width, self.height))

        # Tile position
        if self.column < 0 or self.row < 0:
            metaTable.getEntry("Position").delete()
        else:
            metaTable.getEntry("Position").setNumberArray((self.column, self.row))

        # Custom properties
        if self.getProperties():
            propTable = metaTable.getSubTable("Properties")
            for name, value in self.getProperties().items():
                propTable.getEntry(name).setValue(value)
        self.metadataDirty = False

    # ShuffleboardValue interface

    def buildInto(self, parentTable, metaTable) -> None:
        """Builds the entries for this value.

        :param parentTable: the table containing all the data for the parent. Values that require a
                            complex entry or table structure should call
                            ``parentTable.getSubTable(getTitle())`` to get the table to put data into.
                            Values that only use a single entry should call
                            ``parentTable.getEntry(getTitle())`` to get that entry.

        :param metaTable: the table containing all the metadata for this value and its sub-values
        """
        raise NotImplementedError



from .shuffleboard import Shuffleboard
from .builtinlayouts import BuiltInLayouts
from .builtinwidgets import BuiltInWidgets
from .complexwidget import ComplexWidget
from .component import ShuffleboardComponent
from .container import ShuffleboardContainer
from .eventimportance import EventImportance
from .layout import ShuffleboardLayout
from .simplewidget import SimpleWidget
from .tab import ShuffleboardTab
from .widget import ShuffleboardWidget



# validated: 2019-01-05 TW 01d13220660c edu/wpi/first/wpilibj/shuffleboard/ContainerHelper.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from typing import TYPE_CHECKING, List, Optional, Union, overload

from networktables.entry import NetworkTableEntry
from .builtinlayouts import BuiltInLayouts
from ..sendable import Sendable

if TYPE_CHECKING:
    from .complexwidget import ComplexWidget
    from .simplewidget import SimpleWidget

__all__ = ["ShuffleboardContainer"]


class ShuffleboardContainer:
    """Common interface for objects that can contain shuffleboard components."""

    def __init__(self) -> None:
        self.usedTitles = set()
        self.components = []  # type: List[ShuffleboardContainer]
        self.layouts = {}

    def getComponents(self) -> "List[ShuffleboardContainer]":
        """Gets the components that are direct children of this container."""
        return self.components

    def getLayout(self, title: str, type: Optional[Union[BuiltInLayouts, str]] = None):
        """
        Gets the layout with the given type and title, creating it if it does
        not already exist at the time this method is called.

        :param title: the title of the layout
        :param type:  the type of the layout, eg "List Layout" or "Grid Layout"
        :returns: the layout
        """
        from .layout import ShuffleboardLayout

        if title not in self.layouts:
            if type is None:
                raise KeyError("No layout has been defined with the title '%s'" % title)
            if isinstance(type, BuiltInLayouts):
                type = type.value
            layout = ShuffleboardLayout(self, type, title)
            self.components.append(layout)
            self.layouts[title] = layout

        return self.layouts[title]

    @overload
    def add(self, value: Sendable, *, title: Optional[str] = None) -> "ComplexWidget":
        ...

    @overload
    def add(self, *, title: str, value) -> "SimpleWidget":
        ...

    def add(self, value, *, title: Optional[str] = None):
        """
        Adds a widget to this container to display the given sendable.

        :param value: the Sendable to display, or the default value of the widget
        :param title: the title of the widget (defaults to the
                      Sendable's name if the value is Sendable)
        :returns: a widget to display the sendable data
        :rtype: ComplexWidget or SimpleWidget
        :raises ValueError: if a widget already exists in this container
                            with the given title.
        """
        if isinstance(value, Sendable):
            from .complexwidget import ComplexWidget

            if not title:
                title = value.getName()
                if not title:
                    raise ValueError("Sendable must have a name")
            self._checkTitle(title)
            widget = ComplexWidget(self, title, value)
            self.components.append(widget)
        else:
            from .simplewidget import SimpleWidget

            if not title:
                raise ValueError("A simple widget must have a title")

            self._checkTitle(title)
            self._checkNtType(value)

            widget = SimpleWidget(self, title)
            self.components.append(widget)
            widget.getEntry().setDefaultValue(value)

        return widget

    def _checkNtType(self, data) -> None:
        if not NetworkTableEntry.isValidDataType(data):
            raise TypeError("Cannot add data of type %r to Shuffleboard" % type(data))

    def _checkTitle(self, title: str) -> None:
        if title in self.usedTitles:
            raise ValueError("Title is already in use: " + title)
        self.usedTitles.add(title)

    def addPersistent(self, title: str, defaultValue):
        """
        Adds a widget to this container to display a simple piece of data.
        Unlike :meth:`.add`, the value in the widget will be saved on the
        robot and will be used when the robot program next starts rather than
        `defaultValue`.

        :param title:        the title of the widget
        :param defaultValue: the default value of the widget. note empty lists cannot be used here
        :returns: a widget to display the sendable data
        :rtype: SimpleWidget
        :raises ValueError: if a widget already exists in this container
                            with the given title.

        .. seealso:: :meth:`.add`
        """
        widget = self.add(title=title, value=defaultValue)
        widget.getEntry().setPersistent()
        return widget



# validated: 2018-01-04 TW e1195e8b9dab edu/wpi/first/wpilibj/RobotState.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016-2018. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from .driverstation import DriverStation

__all__ = ["RobotState"]


class RobotState:
    """Provides an interface to determine the current operating state of the
    robot code.
    """

    @staticmethod
    def isDisabled() -> bool:
        return DriverStation.getInstance().isDisabled()

    @staticmethod
    def isEnabled() -> bool:
        return DriverStation.getInstance().isEnabled()

    @staticmethod
    def isOperatorControl() -> bool:
        return DriverStation.getInstance().isOperatorControl()

    @staticmethod
    def isAutonomous() -> bool:
        return DriverStation.getInstance().isAutonomous()

    @staticmethod
    def isTest() -> bool:
        return DriverStation.getInstance().isTest()



# validated: 2019-01-05 TW 01d13220660c edu/wpi/first/wpilibj/shuffleboard/ShuffleboardTab.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from .container import ShuffleboardContainer
from .root import ShuffleboardRoot


class ShuffleboardTab(ShuffleboardContainer):
    def __init__(self, root: ShuffleboardRoot, title: str):
        super().__init__()
        self.root = root
        self.title = title

    def getTitle(self) -> str:
        return self.title

    def getRoot(self) -> ShuffleboardRoot:
        return self.root

    def buildInto(self, parentTable, metaTable) -> None:
        tabTable = parentTable.getSubTable(self.title)
        tabTable.getEntry(".type").setString("ShuffleboardTab")
        for component in self.getComponents():
            component.buildInto(tabTable, metaTable.getSubTable(component.getTitle()))



# validated: 2018-10-30 EN 8b5dc53cc7cd edu/wpi/first/wpilibj/command/InstantCommand.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from .command import Command

__all__ = ["InstantCommand"]


class InstantCommand(Command):
    """
    This command will execute once, then finish immediately afterward.

    Subclassing :class:`.InstantCommand` is shorthand for returning true from
    :meth:`.Command.isFinished`
    """

    def __init__(self, name=None, subsystem=None, func=None):
        """
        Creates a new InstantCommand

        :param name:        the name for this command
        :param requirement: the subsystem this command requires
        :param func:        the function to run when :meth:`.Command.initialize` is run
        """
        super().__init__(name, subsystem=subsystem)
        self.func = func

    def isFinished(self):
        return True

    def _initialize(self):
        """
        Trigger the stored function.

        Called just before this Command runs the first time.
        """
        super()._initialize()
        if self.func is not None:
            self.func()



# validated: 2019-01-05 TW 01d13220660c edu/wpi/first/wpilibj/shuffleboard/ShuffleboardWidget.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from typing import Union

from .builtinwidgets import BuiltInWidgets
from .component import ShuffleboardComponent
from .container import ShuffleboardContainer


class ShuffleboardWidget(ShuffleboardComponent):
    """
    Abstract superclass for widgets.
    """

    def __init__(self, parent: ShuffleboardContainer, title: str):
        super().__init__(parent, title)

    def withWidget(
        self, widgetType: Union[BuiltInWidgets, str]
    ) -> "ShuffleboardWidget":
        """
        Sets the type of widget used to display the data. If not set, the
        default widget type will be used.

        :param widgetType: the type of the widget used to display the data
        :returns: this widget object
        """
        if isinstance(widgetType, BuiltInWidgets):
            self.setType(widgetType.value)
        else:
            self.setType(widgetType)
        return self



# validated: 2017-11-21 EN e1195e8b9dab edu/wpi/first/wpilibj/filters/Filter.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2015-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from .interfaces.pidsource import PIDSource

__all__ = ["Filter"]


class Filter:
    """Superclass for filters"""

    def __init__(self, source: PIDSource) -> None:
        """Constructor.
        
        :param source:
        """
        self.source = PIDSource.from_obj_or_callable(source)

    def setPIDSourceType(self, pidSourceType: PIDSource.PIDSourceType) -> None:
        self.source.setPIDSourceType(pidSourceType)

    def getPIDSourceType(self) -> PIDSource.PIDSourceType:
        return self.source.getPIDSourceType()

    def pidGet(self) -> float:
        raise NotImplementedError

    def get(self) -> float:
        """Returns the current filter estimate without also inserting new data as
        :meth:`pidGet` would do.
        
        :returns: The current filter estimate
        """
        raise NotImplementedError

    def reset(self) -> None:
        """Reset the filter state"""
        raise NotImplementedError

    def pidGetSource(self) -> float:
        """Calls PIDGet() of source
        
        :returns: Current value of source
        """
        return self.source.pidGet()



"""
    Objects in this package allow you to implement a robot using Command-based
    programming.  Command based programming is a design pattern to help you
    organize your robot programs, by organizing your robot program into
    components based on :class:`.Command` and :class:`.Subsystem`
    
    The Python implementation of the Command framework closely follows the
    Java language implementation. RobotPy has several examples of command
    based robots available.
    
    Each one of the objects in the Command framework has detailed
    documentation available. If you need more information, for examples,
    tutorials, and other detailed information on programming your robot
    using this pattern, we recommend that you consult the Java version of the
    `FRC Control System documentation <http://wpilib.screenstepslive.com/s/4485/m/13809/c/88893>`_
"""

from .command import Command
from .commandgroup import CommandGroup
from .conditionalcommand import ConditionalCommand
from .instantcommand import InstantCommand
from .pidcommand import PIDCommand
from .pidsubsystem import PIDSubsystem
from .printcommand import PrintCommand
from .scheduler import Scheduler
from .startcommand import StartCommand
from .subsystem import Subsystem
from .timedcommand import TimedCommand
from .waitcommand import WaitCommand
from .waitforchildren import WaitForChildren
from .waituntilcommand import WaitUntilCommand



# novalidate
"""
    An implementation of java.util.TimerTask which is needed for
    the PID Controller
"""

import threading
from ..timer import Timer

import logging


class TimerTask(threading.Thread):
    def __init__(self, name, period, task_fn):
        super().__init__(name=name, daemon=True)

        if period <= 0:
            raise ValueError("Invalid period %s" % period)

        self.logger = logging.getLogger("wpilib.%s" % name)
        self.period = period
        self.task_fn = task_fn

        self.last_warning = -10

        self.stopped = False

    def cancel(self):
        self.stopped = True
        self.join()

    def run(self):

        self.logger.info("Timer task started")

        try:
            period = self.period
            wait_til = Timer.getFPGATimestamp() + period

            while not self.stopped:

                now = Timer.getFPGATimestamp()
                delay = wait_til - now

                if delay > 0:
                    Timer.delay(delay)
                else:
                    wait_til = now
                    # only emit warning once per second
                    if now - self.last_warning > 1:
                        self.last_warning = now
                        self.logger.warning(
                            "Delay time was negative (%02f: too much going on?)", delay
                        )

                if self.stopped:
                    break

                self.task_fn()

                wait_til += period
        finally:
            self.logger.info("Timer task exited")



# validated: 2017-09-24 AA e1195e8b9dab edu/wpi/first/wpilibj/PIDSource.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import enum
from typing import Callable

__all__ = ["PIDSource"]


class PIDSource:
    """This interface allows for :class:`.PIDController` to automatically read from this
    object.
    """

    @staticmethod
    def from_obj_or_callable(objc: "PIDSource") -> "PIDSource":
        """
            Utility method that gets a PIDSource object
        
            :param objc: An object that implements the PIDSource interface, or
                         a callable
                         
            :returns: an object that implements the PIDSource interface 
        """

        # This isn't pythonic, but will cause errors to happen earlier
        if not hasattr(objc, "pidGet"):
            if not callable(objc):
                raise TypeError("%s must be callable or ")
            return _PIDSourceWrapper(objc)

        if not hasattr(objc, "getPIDSourceType"):
            raise TypeError("%s does not have getPIDSourceType method" % objc)

        return objc

    class PIDSourceType(enum.IntEnum):
        """A description for the type of output value to provide to a
        :class:`.PIDController`"""

        kDisplacement = 0
        kRate = 1

    def setPIDSourceType(self, pidSource: PIDSourceType) -> None:
        """Set which parameter of the device you are using as a process control
        variable.
        
        :param pidSource: An enum to select the parameter.
        """
        raise NotImplementedError

    def getPIDSourceType(self) -> PIDSourceType:
        """Get which parameter of the device you are using as a process control
           variable.
           
        :returns: the currently selected PID source parameter
        """
        raise NotImplementedError

    def pidGet(self) -> float:
        """Get the result to use in :class:`.PIDController`
        
        :returns: the result to use in PIDController
        """
        raise NotImplementedError


class _PIDSourceWrapper(PIDSource):
    def __init__(self, fn: Callable) -> None:
        self.pidGet = fn

    def getPIDSourceType(self) -> PIDSource.PIDSourceType:
        return self.PIDSourceType.kDisplacement



# validated: 2017-09-27 AA e1195e8b9dab edu/wpi/first/wpilibj/interfaces/Potentiometer.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from .pidsource import PIDSource

__all__ = ["Potentiometer"]


class Potentiometer(PIDSource):
    """Interface for a Potentiometer."""

    def get(self) -> float:
        raise NotImplementedError



# validated: 2018-12-15 EN 45f4472d4224 edu/wpi/first/wpilibj/shuffleboard/RecordingController.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from ..driverstation import DriverStation
from networktables.instance import NetworkTablesInstance
from .eventimportance import EventImportance


__all__ = ["RecordingController"]


class RecordingController:
    """Controls Shuffleboard recordings via NetworkTables."""

    kRecordingTableName = "/Shuffleboard/.recording/"
    kRecordingControlKey = kRecordingTableName + "RecordData"
    kRecordingFileNameFormatKey = kRecordingTableName + "FileNameFormat"
    kEventMarkerTableName = kRecordingTableName + "events"

    def __init__(self, ntInstance: NetworkTablesInstance) -> None:
        self.recordingControlEntry = ntInstance.getEntry(
            RecordingController.kRecordingControlKey
        )
        self.recordingFileNameFormatEntry = ntInstance.getEntry(
            RecordingController.kRecordingFileNameFormatKey
        )
        self.eventsTable = ntInstance.getTable(
            RecordingController.kEventMarkerTableName
        )

    def startRecording(self) -> None:
        self.recordingControlEntry.setBoolean(True)

    def stopRecording(self) -> None:
        self.recordingControlEntry.setBoolean(False)

    def setRecordingFileNameFormat(self, format: str) -> None:
        self.recordingFileNameFormatEntry.setString(format)

    def clearRecordingFileNameFormat(self) -> None:
        self.recordingFileNameFormatEntry.delete()

    def addEventMarker(self, name: str, description: str, importance: EventImportance):
        if not name or name.isspace():
            DriverStation.reportError("Shuffleboard event name was not specified", True)
            return

        eventDescription = description

        if description is None or description.isspace():
            eventDescription = ""

        self.eventsTable.getSubTable(name).getEntry("Info").setStringArray(
            [eventDescription, importance.value]
        )



# validated: 2018-11-18 EN 0614913f1abb edu/wpi/first/wpilibj/PWMSpeedController.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from .pwm import PWM
from .sendablebuilder import SendableBuilder

__all__ = ["PWMSpeedController"]


class PWMSpeedController(PWM):
    """
        Common base class for all PWM Speed Controllers.
    """

    def __init__(self, channel: int) -> None:
        super().__init__(channel)
        self.isInverted = False

    def getDescription(self):
        return "PWM %" % (self.getChannel(),)

    def set(self, speed: float) -> None:
        """Set the PWM value.

        The PWM value is set using a range of -1.0 to 1.0, appropriately
        scaling the value for the FPGA.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        """
        self.setSpeed(-speed if self.isInverted else speed)
        self.feed()

    def setInverted(self, isInverted: bool) -> None:
        """
        Common interface for inverting the direction of a speed controller.

        :param isInverted: The state of inversion (True is inverted).
        """
        self.isInverted = isInverted

    def getInverted(self) -> bool:
        """
        Common interface for inverting the direction of a speed controller.

        :returns: The state of inversion (True is inverted)
        """
        return self.isInverted

    def get(self) -> float:
        """Get the recently set value of the PWM.

        :returns: The most recently set value for the PWM between -1.0 and 1.0.
        """
        return self.getSpeed()

    def disable(self):
        self.setDisabled()

    def pidWrite(self, output: float) -> None:
        """Write out the PID value as seen in the PIDOutput base object.

        :param output: Write out the PWM value as was found in the
            :class:`PIDController`.
        """
        self.set(output)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("Speed Controller")
        builder.setActuator(True)
        builder.setSafeState(self.setDisabled)
        builder.addDoubleProperty("Value", self.getSpeed, self.setSpeed)



# validated: 2019-01-02 EN 89d15f061b33 edu/wpi/first/wpilibj/PIDBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2018. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import threading
import warnings

import hal

from .sendablebuilder import SendableBuilder
from .interfaces.pidsource import PIDSource
from .sendablebase import SendableBase
from .resource import Resource
from .timer import Timer
from .lineardigitalfilter import LinearDigitalFilter

__all__ = ["PIDBase"]


class PIDBase(SendableBase):
    """Class implements a PID Control Loop.

    Creates a separate thread which reads the given PIDSource and takes care of the integral
    calculations, as well as writing the given PIDOutput.

    This feedback controller runs in discrete time, so time deltas are not used in the integral
    and derivative calculations. Therefore, the sample rate affects the controller's behavior for a
    given set of PID constants.

    """

    kDefaultPeriod = 0.05
    instances = 0

    PIDSourceType = PIDSource.PIDSourceType

    # Tolerance is the type of tolerance used to specify if the PID controller
    # is on target.  The various implementations of this such as
    # PercentageTolerance and AbsoluteTolerance specify types of tolerance
    # specifications to use.
    def PercentageTolerance_onTarget(self, percentage: float) -> bool:
        return abs(self.getError()) < percentage / 100.0 * self.inputRange

    def AbsoluteTolerance_onTarget(self, value: float) -> bool:
        return abs(self.getError()) < value

    def __init__(self, Kp, Ki, Kd, Kf, source, output):
        """Allocate a PID object with the given constants for P, I, D, and F

        :param Kp: the proportional coefficient
        :type  Kp: float or int
        :param Ki: the integral coefficient
        :type  Ki: float or int
        :param Kd: the derivative coefficient
        :type  Kd: float or int
        :param Kf: the feed forward term
        :type  Kf: float or int
        :param source: Called to get values
        :type  source: A function, or an object that implements :class:`.PIDSource`
        :param output: Receives the output percentage
        :type  output: A function, or an object that implements :class:`.PIDOutput`
        """

        super().__init__(False)

        self.P = Kp  # factor for "proportional" control
        self.I = Ki  # factor for "integral" control
        self.D = Kd  # factor for "derivative" control
        self.F = Kf  # factor for feedforward term
        self.pidOutput = output
        self.origSource = source
        self.filter = LinearDigitalFilter.movingAverage(self.origSource, 1)
        self.pidInput = self.filter

        self.pidInput = PIDSource.from_obj_or_callable(self.pidInput)

        if hasattr(self.pidOutput, "pidWrite"):
            self.pidOutput = self.pidOutput.pidWrite

        # fmt: off
        self.maximumOutput = 1.0    # |maximum output|
        self.minimumOutput = -1.0   # |minimum output|
        self.maximumInput = 0.0     # maximum input - limit setpoint to this
        self.minimumInput = 0.0     # minimum input - limit setpoint to this
        self.inputRange = 0.0       # input range - difference between maximum and minimum
        self.continuous = False     # do the endpoints wrap around? eg. Absolute encoder
        self.enabled = False        # is the pid controller enabled
        self.prevError = 0.0        # the prior error (used to compute velocity)
        self.totalError = 0.0       #the sum of the errors for use in the integral calc
        self.setpoint = 0.0
        self.prevSetpoint = 0.0
        self.error = 0.0
        self.result = 0.0
        # fmt: on

        self.mutex = threading.RLock()
        self.pidWriteMutex = threading.RLock()

        self.setpointTimer = Timer()
        self.setpointTimer.start()

        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        PIDBase.instances += 1
        hal.report(hal.UsageReporting.kResourceType_PIDController, PIDBase.instances)
        self.setName("PIDController", PIDBase.instances)

    def _calculate(self) -> None:
        """Read the input, calculate the output accordingly, and write to the
        output.  This should only be called by the PIDTask and is created
        during initialization."""
        if self.origSource is None or self.pidOutput is None:
            return

        with self.mutex:
            enabled = self.enabled  # take snapshot of these values...

        if enabled:
            feedForward = self.calculateFeedForward()

            with self.mutex:
                input = self.pidInput.pidGet()
                pidSourceType = self.pidInput.getPIDSourceType()
                P = self.P
                I = self.I
                D = self.D
                minimumOutput = self.minimumOutput
                maximumOutput = self.maximumOutput
                prevError = self.prevError
                error = self.getContinuousError(self.setpoint - input)
                totalError = self.totalError

            # start

            if pidSourceType == self.PIDSourceType.kRate:
                if P != 0:
                    totalError = self.clamp(
                        totalError + error, minimumOutput / P, maximumOutput / P
                    )

                result = P * totalError + D * error + feedForward

            else:
                if I != 0:
                    totalError = self.clamp(
                        totalError + error, minimumOutput / I, maximumOutput / I
                    )

                result = (
                    P * error + I * totalError + D * (error - prevError) + feedForward
                )

            result = self.clamp(result, minimumOutput, maximumOutput)

            with self.pidWriteMutex:
                with self.mutex:
                    enabled = self.enabled
                if enabled:
                    self.pidOutput(result)

            with self.mutex:
                self.prevError = error
                self.error = error
                self.totalError = totalError
                self.result = result

    def calculateFeedForward(self) -> float:
        """Calculate the feed forward term

        Both of the provided feed forward calculations are velocity feed forwards.
        If a different feed forward calculation is desired, the user can override
        this function and provide his or her own. This function  does no
        synchronization because the PIDController class only calls it in
        synchronized code, so be careful if calling it oneself.

        If a velocity PID controller is being used, the F term should be set to 1
        over the maximum setpoint for the output. If a position PID controller is
        being used, the F term should be set to 1 over the maximum speed for the
        output measured in setpoint units per this controller's update period (see
        the default period in this class's constructor).
        """
        if self.pidInput.getPIDSourceType() == self.PIDSourceType.kRate:
            return self.F * self.getSetpoint()
        else:
            temp = self.F * self.getDeltaSetpoint()
            self.prevSetpoint = self.setpoint
            self.setpointTimer.reset()
            return temp

    def setPID(self, p: float, i: float, d: float, f: float = 0.0) -> None:
        """Set the PID Controller gain parameters.
        Set the proportional, integral, and differential coefficients.

        :param p: Proportional coefficient
        :param i: Integral coefficient
        :param d: Differential coefficient
        :param f: Feed forward coefficient (optional, default is 0.0)
        """
        with self.mutex:
            self.P = p
            self.I = i
            self.D = d
            self.F = f

    def setP(self, p) -> None:
        """
        Set the Proportional coefficient of the PID controller gain.

        :param p: Proportional coefficient
        """
        with self.mutex:
            self.P = p

    def setI(self, i) -> None:
        """
        Set the Integral coefficient of the PID controller gain.

        :param i: Integral coefficient
        """
        with self.mutex:
            self.I = i

    def setD(self, d) -> None:
        """
        Set the Differential coefficient of the PID controller gain.

        :param d: differential coefficient
        """
        with self.mutex:
            self.D = d

    def setF(self, f) -> None:
        """
        Set the Feed forward coefficient of the PID controller gain.

        :param f: feed forward coefficient
        """
        with self.mutex:
            self.F = f

    def getP(self) -> float:
        """Get the Proportional coefficient.

        :returns: proportional coefficient
        """
        with self.mutex:
            return self.P

    def getI(self) -> float:
        """Get the Integral coefficient

        :returns: integral coefficient
        """
        with self.mutex:
            return self.I

    def getD(self) -> float:
        """Get the Differential coefficient.

        :returns: differential coefficient
        """
        with self.mutex:
            return self.D

    def getF(self) -> float:
        """Get the Feed forward coefficient.

        :returns: feed forward coefficient
        """
        with self.mutex:
            return self.F

    def get(self) -> float:
        """Return the current PID result.
        This is always centered on zero and constrained the the max and min
        outs.

        :returns: the latest calculated output
        """
        with self.mutex:
            return self.result

    def setContinuous(self, continuous: bool = True) -> None:
        """Set the PID controller to consider the input to be continuous.
        Rather then using the max and min input range as constraints, it considers them
        to be the same point and automatically calculates the shortest route
        to the setpoint.

        :param continuous: Set to True turns on continuous, False turns off
            continuous
        """
        if continuous and self.inputRange <= 0:
            raise ValueError("No input range set when calling setContinuous().")
        with self.mutex:
            self.continuous = continuous

    def setInputRange(self, minimumInput: float, maximumInput: float) -> None:
        """Sets the maximum and minimum values expected from the input.

        :param minimumInput: the minimum percentage expected from the input
        :param maximumInput: the maximum percentage expected from the output
        """
        with self.mutex:
            if minimumInput > maximumInput:
                raise ValueError("Lower bound is greater than upper bound")
            self.minimumInput = minimumInput
            self.maximumInput = maximumInput
            self.inputRange = self.maximumInput - self.minimumInput
            self.setSetpoint(self.setpoint)

    def setOutputRange(self, minimumOutput: float, maximumOutput: float) -> None:
        """Sets the minimum and maximum values to write.

        :param minimumOutput: the minimum percentage to write to the output
        :param maximumOutput: the maximum percentage to write to the output
        """
        with self.mutex:
            if minimumOutput > maximumOutput:
                raise ValueError("Lower bound is greater than upper bound")
            self.minimumOutput = minimumOutput
            self.maximumOutput = maximumOutput

    def setSetpoint(self, setpoint: float) -> None:
        """Set the setpoint for the PIDController.

        :param setpoint: the desired setpoint
        """
        with self.mutex:
            if self.maximumInput > self.minimumInput:
                if setpoint > self.maximumInput:
                    newsetpoint = self.maximumInput
                elif setpoint < self.minimumInput:
                    newsetpoint = self.minimumInput
                else:
                    newsetpoint = setpoint
            else:
                newsetpoint = setpoint
            self.setpoint = newsetpoint

    def getSetpoint(self) -> float:
        """Returns the current setpoint of the PIDController.

        :returns: the current setpoint
        """
        with self.mutex:
            return self.setpoint

    def getDeltaSetpoint(self) -> float:
        """Returns the change in setpoint over time of the PIDController

        :returns: the change in setpoint over time
        """
        with self.mutex:
            t = self.setpointTimer.get()
            # During testing/simulation it is possible to get a divide by zero because
            # the threads' calls aren't strictly aligned with the master clock.
            if t:
                return (self.setpoint - self.prevSetpoint) / t
            else:
                return 0.0

    def getError(self) -> float:
        """Returns the current difference of the input from the setpoint.

        :return: the current error
        """
        with self.mutex:
            return self.getContinuousError(self.getSetpoint() - self.pidInput.pidGet())

    def getAvgError(self) -> float:
        """
        Returns the current difference of the error over the past few iterations. You can specify the
        number of iterations to average with setToleranceBuffer() (defaults to 1). getAvgError() is
        used for the onTarget() function.

        .. deprecated :: 2018.0.0
            Use getError(), which is now already filtered.

        :returns: the current average of the error
        """
        warnings.warn("use getRate instead", DeprecationWarning)
        with self.mutex:
            return self.getError()

    def setPIDSourceType(self, pidSourceType: PIDSourceType) -> None:
        """Sets what type of input the PID controller will use

        :param pidSourceType: the type of input
        """
        self.pidInput.setPIDSourceType(pidSourceType)

    def getPIDSourceType(self) -> PIDSourceType:
        """Returns the type of input the PID controller is using

        :returns: the PID controller input type
        """
        return self.pidInput.getPIDSourceType()

    def setAbsoluteTolerance(self, absvalue: float) -> None:
        """Set the absolute error which is considered tolerable for use with
        :func:`onTarget`.

        :param absvalue: absolute error which is tolerable in the units of the
            input object
        """
        with self.mutex:
            self.onTarget = lambda: self.AbsoluteTolerance_onTarget(absvalue)

    def setPercentTolerance(self, percentage: float) -> None:
        """Set the percentage error which is considered tolerable for use with
        :func:`onTarget`. (Input of 15.0 = 15 percent)

        :param percentage: percent error which is tolerable
        """
        with self.mutex:
            self.onTarget = lambda: self.PercentageTolerance_onTarget(percentage)

    def setToleranceBuffer(self, bufLength: int) -> None:
        """Set the number of previous error samples to average for tolerancing. When
        determining whether a mechanism is on target, the user may want to use a
        rolling average of previous measurements instead of a precise position or
        velocity. This is useful for noisy sensors which return a few erroneous
        measurements when the mechanism is on target. However, the mechanism will
        not register as on target for at least the specified bufLength cycles.

        :param bufLength: Number of previous cycles to average.
        :type bufLength: int

        .. deprecated:: 2018.0.0
            Use a LinearDigitalFilter as the input
        """
        with self.mutex:
            self.filter = LinearDigitalFilter.movingAverage(self.origSource, bufLength)
            self.pidInput = self.filter

    def onTarget(self) -> bool:
        """Return True if the error is within the percentage of the total input
        range, determined by setTolerance. This assumes that the maximum and
        minimum input were set using :func:`setInput`.

        :returns: True if the error is less than the tolerance
        """
        with self.mutex:
            # python-specific: this is equivalent to the NullTolerance object in Java
            raise ValueError("No tolerance value set when calling onTarget().")

    def reset(self) -> None:
        """Reset the previous error, the integral term, and disable the controller.

        """
        with self.mutex:
            self.prevError = 0
            self.totalError = 0
            self.result = 0

    def pidWrite(self, output: float) -> None:
        """Passes the output directly to setSetpoint().

        PIDControllers can be nested by passing a PIDController as another PIDController's output.
        In that case, the output of the parent controller becomes the input (i.e., the reference) of
        the child.

        It is the caller's responsibility to put the data into a valid form for setSetpoint().

        """
        self.setSetpoint(output)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("PIDController")
        builder.setSafeState(self.reset)
        builder.addDoubleProperty("p", self.getP, self.setP)
        builder.addDoubleProperty("i", self.getI, self.setI)
        builder.addDoubleProperty("d", self.getD, self.setD)
        builder.addDoubleProperty("f", self.getF, self.setF)
        builder.addDoubleProperty("setpoint", self.getSetpoint, self.setSetpoint)

    def getContinuousError(self, error: float) -> float:
        """
        Wraps error around for continuous inputs. The original error is
        returned if continuous mode is disabled. This is an unsynchronized
        function.

        :param error: The current error of the PID controller.
        :return: Error for continuous inputs.
        """
        if self.continuous and self.inputRange > 0:
            error %= self.inputRange
            if abs(error) > self.inputRange / 2:
                if error > 0:
                    return error - self.inputRange
                else:
                    return error + self.inputRange
        return error

    def clamp(self, value: float, low: float, high: float) -> float:
        return max(low, min(value, high))



# validated: 2017-12-15 EN f9bece2ffbf7 edu/wpi/first/wpilibj/PWMTalonSRX.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal

from .pwmspeedcontroller import PWMSpeedController

__all__ = ["PWMTalonSRX"]


class PWMTalonSRX(PWMSpeedController):
    """
    Cross the Road Electronics (CTRE) Talon SRX Speed Controller with PWM control.

    .. seealso:: See :class:`ctre.cantalon.CANTalon` for CAN control of Talon SRX.
    """

    def __init__(self, channel: int) -> None:
        """Constructor for a TalonSRX connected via PWM.

        :param channel: The PWM channel that the PWMTalonSRX is attached to. 0-9 are on-board, 10-19 are on the MXP port.

        .. note ::

            The PWMTalonSRX uses the following bounds for PWM values. These values
            should work reasonably well for most controllers, but if users
            experience issues such as asymmetric behavior around the deadband
            or inability to saturate the controller in either direction,
            calibration is recommended.  The calibration procedure can be
            found in the TalonSRX User Manual available from CTRE.

            - 2.004ms = full "forward"
            - 1.520ms = the "high end" of the deadband range
            - 1.500ms = center of the deadband range (off)
            - 1.480ms = the "low end" of the deadband range
            - 0.997ms = full "reverse"
        """
        super().__init__(channel)
        self.setBounds(2.004, 1.52, 1.50, 1.48, 0.997)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setSpeed(0)
        self.setZeroLatch()

        hal.report(hal.UsageReporting.kResourceType_PWMTalonSRX, self.getChannel())
        self.setName("PWMTalonSRX", self.getChannel())



"""
    This package contains objects that can be used to determine the
    requirements of various interfaces used in WPILib. 
    
    Generally, the python version of WPILib does not require that you inherit
    from any of these interfaces, but instead will allow you to use custom 
    objects as long as they have the same methods.
"""

from .accelerometer import Accelerometer
from .controller import Controller
from .counterbase import CounterBase
from .generichid import GenericHID
from .gyro import Gyro
from .namedsendable import NamedSendable
from .pidoutput import PIDOutput
from .pidsource import PIDSource
from .potentiometer import Potentiometer
from .speedcontroller import SpeedController



# validated: 2019-01-04 TW 97ba195b881e edu/wpi/first/wpilibj/smartdashboard/SendableBuilderImpl.java
# ----------------------------------------------------------------------------
# Copyright (c) 2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from typing import Callable, Any, List, Optional
from networktables import NetworkTables
from networktables.entry import NetworkTableEntry
from networktables.networktable import NetworkTable

__all__ = ["SendableBuilder"]


class Property:
    def __init__(
        self,
        table: NetworkTable,
        key: str,
        update: Callable,
        setter: Callable,
        listen_local: bool,
    ) -> None:
        self.entry = table.getEntry(key)
        self.key = key
        self.update = update
        self.setter = setter
        self.listener = None
        # python-specific: sendable chooser needs to be able to hear local updates
        self.listen_local = listen_local

    def createListener(self, entry: NetworkTableEntry) -> None:
        flags = (
            NetworkTables.NotifyFlags.IMMEDIATE
            | NetworkTables.NotifyFlags.NEW
            | NetworkTables.NotifyFlags.UPDATE
        )
        if self.listen_local:
            flags |= NetworkTables.NotifyFlags.LOCAL
        self.listener = entry.addListener(
            lambda entry, key, value, param: self.setter(value), flags
        )

    def startListener(self) -> None:
        if self.listener is None and self.setter is not None:
            self.createListener(self.entry)

    def stopListener(self) -> None:
        if self.listener is not None:
            self.entry.removeListener(self.listener)
            self.listener = None


class SendableBuilder:
    def __init__(self) -> None:
        self.table = None
        self._updateTable = None
        self.safeState = None
        self.properties = []
        self.controllableEntry = None
        self.actuator = False

    def setTable(self, table: NetworkTable) -> None:
        """
        Set the network table.  Must be called prior to any Add* functions being called.

        :param table: Network table
        """
        self.table = table
        self.controllableEntry = table.getEntry(".controllable")

    def getTable(self) -> NetworkTable:
        """
        Get the network table.

        :returns: The network table
        """
        return self.table

    def isActuator(self) -> bool:
        """
        Return whether this sendable should be treated as an actuator.
        """
        return self.actuator

    def updateTable(self) -> None:
        """
        Update the network table values by calling the getters for all properties.
        """
        for prop in self.properties:
            if prop.update is not None:
                prop.update(prop.entry)

        if self._updateTable is not None:
            self._updateTable()

    def startListeners(self) -> None:
        """Hook setters for all properties"""
        for prop in self.properties:
            prop.startListener()
        if self.controllableEntry is not None:
            self.controllableEntry.setBoolean(True)

    def stopListeners(self) -> None:
        """Unhook setters for all properties"""
        for prop in self.properties:
            prop.stopListener()
        if self.controllableEntry is not None:
            self.controllableEntry.setBoolean(False)

    def startLiveWindowMode(self) -> None:
        """
        Start LiveWindow mode by hooking the setters for all properties. Also calls
        the safeState function if one was provided.
        """
        if self.safeState is not None:
            self.safeState()

        self.startListeners()

    def stopLiveWindowMode(self) -> None:
        """
        Stop LiveWindow mode by unhooking the setters for all properties. Also calls
        the safeState function if one was provided.
        """
        self.stopListeners()
        if self.safeState is not None:
            self.safeState()

    def setSmartDashboardType(self, type: str) -> None:
        """
        Set the string representation of the named data type that will be used
        by the smart dashboard for this sendable.
   
        :param type: data type
        """
        self.table.getEntry(".type").setString(type)

    def setActuator(self, value) -> None:
        """
        Set a flag indicating if this sendable should be treated as an actuator.
        By default this flag is false.

        :param value: true if actuator, false if not
        """
        self.table.getEntry(".actuator").setBoolean(value)
        self.actuator = value

    def setSafeState(self, func: Callable) -> None:
        """
        Set the function that should be called to set the Sendable into a safe
        state.  This is called when entering and exiting Live Window mode.

        :param func:    function
        """
        self.safeState = func

    def setUpdateTable(self, func: Callable) -> None:
        """
        Set the function that should be called to update the network table
        for things other than properties.  Note this function is not passed
        the network table object; instead it should use the entry handles
        returned by getEntry().

        :param func:    function
        """
        self._updateTable = func

    def getEntry(self, key: str) -> NetworkTableEntry:
        """
        Add a property without getters or setters.  This can be used to get
        entry handles for the function called by setUpdateTable().

        :param key:   property name
        :returns: Network table entry
        """
        return self.table.getEntry(key)

    def _addProperty(
        self, key: str, updater: Callable, setter: Callable, listen_local: bool
    ) -> None:
        prop = Property(self.table, key, updater, setter, listen_local)
        self.properties.append(prop)

    def addBooleanProperty(
        self,
        key: str,
        getter: Optional[Callable[[], bool]],
        setter: Optional[Callable[[bool], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a boolean property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = None if getter is None else lambda entry: entry.setBoolean(getter())
        self._addProperty(key, updater, setter, local)

    def addDoubleProperty(
        self,
        key: str,
        getter: Optional[Callable[[], float]],
        setter: Optional[Callable[[float], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a double property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = None if getter is None else lambda entry: entry.setDouble(getter())
        self._addProperty(key, updater, setter, local)

    def addStringProperty(
        self,
        key: str,
        getter: Optional[Callable[[], str]],
        setter: Optional[Callable[[str], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a string property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = None if getter is None else lambda entry: entry.setString(getter())
        self._addProperty(key, updater, setter, local)

    def addBooleanArrayProperty(
        self,
        key: str,
        getter: Optional[Callable[[], List[bool]]],
        setter: Optional[Callable[[List[bool]], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a boolean array property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = (
            None if getter is None else lambda entry: entry.setBooleanArray(getter())
        )
        self._addProperty(key, updater, setter, local)

    def addDoubleArrayProperty(
        self,
        key: str,
        getter: Optional[Callable[[], List[float]]],
        setter: Optional[Callable[[List[float]], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a double array property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = (
            None if getter is None else lambda entry: entry.setDoubleArray(getter())
        )
        self._addProperty(key, updater, setter, local)

    def addStringArrayProperty(
        self,
        key: str,
        getter: Optional[Callable[[], List[str]]],
        setter: Optional[Callable[[List[str]], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a string array property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = (
            None if getter is None else lambda entry: entry.setStringArray(getter())
        )
        self._addProperty(key, updater, setter, local)

    def addRawProperty(
        self,
        key: str,
        getter: Optional[Callable[[], bytes]],
        setter: Optional[Callable[[bytes], Any]],
        local: bool = False,
    ) -> None:
        """
        Add a raw property.
   
        :param key:     property name
        :param getter:  getter function (returns current value)
        :param setter:  setter function (sets new value)
        :param local:   (python-specific) if True, setter will be called on local updates
        """
        updater = None if getter is None else lambda entry: entry.setRaw(getter())
        self._addProperty(key, updater, setter, local)



# validated: 2019-01-04 TW 0d7d880261b6 edu/wpi/first/wpilibj/smartdashboard/SendableChooser.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import threading
import warnings
from typing import Any

from .sendablebase import SendableBase
from .sendablebuilder import SendableBuilder

__all__ = ["SendableChooser"]


class SendableChooser(SendableBase):
    """A useful tool for presenting a selection of options to be displayed on
    the :class:`.SmartDashboard`

    For instance, you may wish to be able to select between multiple
    autonomous modes. You can do this by putting every possible :class:`.Command`
    you want to run as an autonomous into a SendableChooser and then put
    it into the :class:`.SmartDashboard` to have a list of options appear on the
    laptop. Once autonomous starts, simply ask the SendableChooser what
    the selected value is.

    Example::

        # This shows the user two options on the SmartDashboard
        chooser = wpilib.SendableChooser()
        chooser.addObject('option1', '1')
        chooser.addObject('option2', '2')

        wpilib.SmartDashboard.putData('Choice', chooser)

        # .. later, ask to see what the user selected?
        value = chooser.getSelected()

    """

    # The key for the default value
    DEFAULT = "default"
    # The key for the selected option
    SELECTED = "selected"
    # The key for the active option
    ACTIVE = "active"
    # The key for the option array
    OPTIONS = "options"
    # The key for the instance number
    INSTANCE = ".instance"

    _increment_lock = threading.Lock()
    _instances = 0

    def __init__(self) -> None:
        """Instantiates a SendableChooser.
        """
        super().__init__(addLiveWindow=False)
        self.map = {}
        self.selected = None
        self.defaultChoice = ""
        self.activeEntries = []
        self.mutex = threading.RLock()
        with SendableChooser._increment_lock:
            self.instance = SendableChooser._instances
            SendableChooser._instances += 1

    def addOption(self, name: str, object: Any) -> None:
        """Adds the given object to the list of options. On the
        :class:`.SmartDashboard` on the desktop, the object will appear as the
        given name.

        :param name: the name of the option
        :param object: the option
        """
        self.map[name] = object

    def addObject(self, name: str, object: Any) -> None:
        """Adds the given object to the list of options. On the
        :class:`.SmartDashboard` on the desktop, the object will appear as the
        given name.

        :param name: the name of the option
        :param object: the option

        .. deprecated:: 2019.0.0
            Use :meth:`addOption` instead
        """
        warnings.warn(
            "SendableChooser.addObject is deprecated, use addOption instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.addOption(name, object)

    def setDefaultOption(self, name: str, object: Any) -> None:
        """Add the given object to the list of options and marks it as the
        default.  Functionally, this is very close to :meth:`.addObject` except
        that it will use this as the default option if none other is
        explicitly selected.

        :param name: the name of the option
        :param object: the option
        """
        if name is None:
            raise ValueError("Name cannot be None")
        self.defaultChoice = name
        self.addOption(name, object)

    def addDefault(self, name: str, object: Any) -> None:
        """Add the given object to the list of options and marks it as the
        default.  Functionally, this is very close to :meth:`.addObject` except
        that it will use this as the default option if none other is
        explicitly selected.

        :param name: the name of the option
        :param object: the option

        .. deprecated:: 2019.0.0
            Use :meth:`setDefaultOption` instead
        """
        warnings.warn(
            "SendableChooser.addDefault is deprecated, use setDefaultOption instead",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.setDefaultOption(name, object)

    def getSelected(self) -> Any:
        """Returns the object associated with the selected option. If there
        is none selected, it will return the default. If there is none
        selected and no default, then it will return None.

        :returns: the object associated with the selected option
        """
        with self.mutex:
            if self.selected is not None:
                return self.map.get(self.selected)
            else:
                return self.map.get(self.defaultChoice)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("String Chooser")
        builder.getEntry(SendableChooser.INSTANCE).setDouble(self.instance)
        builder.addStringProperty(self.DEFAULT, lambda: self.defaultChoice, None)
        builder.addStringArrayProperty(self.OPTIONS, lambda: self.map.keys(), None)

        def _active_property_getter():
            with self.mutex:
                return self.selected if self.selected else self.defaultChoice

        builder.addStringProperty(SendableChooser.ACTIVE, _active_property_getter, None)

        with self.mutex:
            self.activeEntries.append(builder.getEntry(SendableChooser.ACTIVE))

        def _selected_property_setter(val):
            with self.mutex:
                self.selected = val
                for entry in self.activeEntries:
                    entry.setString(val)

        # python-specific: set local=True
        builder.addStringProperty(
            SendableChooser.SELECTED, None, _selected_property_setter, local=True
        )



# validated: 2017-11-13 TW 21585f70a88e edu/wpi/first/wpilibj/GamepadBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import warnings
from typing import Optional

from .generichid import GenericHID

__all__ = ["GamepadBase"]


class GamepadBase(GenericHID):
    """
    GamepadBase Interface.

    .. deprecated: 2018.0.0
       Inherit directly from :class:`.GenericHID` instead.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        warnings.warn(
            "GamepadBase is deprecated, inherit directly from GenericHID instead",
            DeprecationWarning,
            stacklevel=2,
        )

    def getRawAxis(self, axis: int) -> float:
        raise NotImplementedError

    def getBumper(self, hand: GenericHID.Hand) -> bool:
        """Is the bumper pressed.

        :param hand: which hand
        :returns: true if the bumper is pressed
        """
        raise NotImplementedError

    def getStickButton(self, hand: Optional[GenericHID.Hand] = None) -> bool:
        raise NotImplementedError

    def getRawButton(self, button: int) -> bool:
        raise NotImplementedError

    def getPOV(self, pov: int = 0) -> int:
        raise NotImplementedError

    def getPOVCount(self) -> int:
        raise NotImplementedError

    def getType(self) -> GenericHID.HIDType:
        raise NotImplementedError

    def getName(self) -> str:
        raise NotImplementedError

    def setOutput(self, outputNumber: int, value: bool) -> None:
        raise NotImplementedError

    def setOutputs(self, value: bool) -> bool:
        raise NotImplementedError

    def setRumble(self, type: GenericHID.RumbleType, value: float) -> bool:
        raise NotImplementedError



# validated: 2018-11-17 EN 8373e0361b2c edu/wpi/first/wpilibj/Controller.java
__all__ = ["Controller"]


class Controller:
    """An interface for controllers. Controllers run control loops, the most
    command are PID controllers and there variants, but this includes anything
    that is controlling an actuator in a separate thread.
    """

    def enable(self) -> None:
        """Allows the control loop to run."""
        raise NotImplementedError

    def disable(self) -> None:
        """Stops the control loop from running until explicitly re-enabled by
        calling :meth:`enable`.
        """
        raise NotImplementedError



# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from collections import namedtuple

__all__ = ["AccumulatorResult"]

AccumulatorResult = namedtuple("AccumulatorResult", ["value", "count"])

AccumulatorResult.__doc__ = "Structure for holding the values stored in an accumulator."
AccumulatorResult.value.__doc__ = "The total value accumulated."
AccumulatorResult.count.__doc__ = "The number of sample value was accumulated over."



# validated: 2019-01-10 DV 01d13220660c edu/wpi/first/wpilibj/shuffleboard/BuiltInWidgets.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import enum


class BuiltInWidgets(enum.Enum):
    """The types of the widgets bundled with Shuffleboard.

    For example, setting a number to be displayed with a slider::

        example_entry = (
            Shuffleboard.getTab("My Tab")
            .add("My Number", 0)
            .withWidget(BuiltInWidgets.kNumberSlider)
            .withProperties({"min": 0, "max": 1})
            .getEntry()
        )

    Each value in this enum goes into detail on what data types that
    widget can support, as well as the custom properties that widget uses.
    """

    #: Displays a value with a simple text field.
    #:
    #: Supported types:
    #:
    #: - String
    #: - Number
    #: - Boolean
    #:
    #: This widget has no custom properties.
    kTextView = "Text View"

    #: Displays a number with a controllable slider.
    #:
    #: Supported types:
    #:
    #: - Number
    #:
    #: Custom properties:
    #:
    #: .. csv-table::
    #:    :header: Name, Type, Default Value, Notes
    #:    :widths: auto
    #:
    #:    Min,             Number, -1.0,   The minimum value of the slider
    #:    Max,             Number, 1.0,    The maximum value of the slider
    #:    Block increment, Number, 0.0625, How much to move the slider by with the arrow keys
    kNumberSlider = "Number Slider"

    #: Displays a number with a view-only bar.
    #:
    #: Supported types:
    #:
    #: - Number
    #:
    #: Custom properties:
    #:
    #: ======  ======  =============  ====================================
    #: Name    Type    Default Value  Notes
    #: ======  ======  =============  ====================================
    #: Min     Number  -1.0           The minimum value of the bar
    #: Max     Number  1.0            The maximum value of the bar
    #: Center  Number  0              The center ("zero") value of the bar
    #: ======  ======  =============  ====================================
    kNumberBar = "Number Bar"

    #: Displays a number with a view-only dial.
    #: Displayed values are rounded to the nearest integer.
    #:
    #: Supported types:
    #:
    #: - Number
    #:
    #: Custom properties:
    #:
    #: .. list-table::
    #:    :widths: auto
    #:    :header-rows: 1
    #:
    #:    * - Name
    #:      - Type
    #:      - Default Value
    #:      - Notes
    #:    * - Min
    #:      - Number
    #:      - 0
    #:      - The minimum value of the dial
    #:    * - Max
    #:      - Number
    #:      - 100
    #:      - The maximum value of the dial
    #:    * - Show value
    #:      - Boolean
    #:      - true
    #:      - Whether or not to show the value as text
    kDial = "Simple Dial"

    #: Displays a number with a graph.
    #:
    #: .. note::
    #:    Graphs can be taxing on the computer running the dashboard.
    #:    Keep the number of visible data points to a minimum.
    #:    Making the widget smaller also helps with performance,
    #:    but may cause the graph to become difficult to read.
    #:
    #: Supported types:
    #:
    #: - Number
    #: - Number array
    #:
    #: Custom properties:
    #:
    #: ============ ====== ============= =====================================================
    #: Name         Type   Default Value Notes
    #: ============ ====== ============= =====================================================
    #: Visible time Number 30            How long, in seconds, should past data be visible for
    #: ============ ====== ============= =====================================================
    kGraph = "Graph"

    #: Displays a boolean value as a large colored box.
    #:
    #: Supported types:
    #:
    #: - Boolean
    #:
    #: Custom properties:
    #:
    #: ================ ===== ============= ===============================================================================
    #: Name             Type  Default Value Notes
    #: ================ ===== ============= ===============================================================================
    #: Color when true  Color "green"       Can be specified as a string (``"#00FF00"``) or a rgba integer (``0x00FF0000``)
    #: Color when false Color "red"         Can be specified as a string or a number
    #: ================ ===== ============= ===============================================================================
    kBooleanBox = "Boolean Box"

    #: Displays a boolean with a large interactive toggle button.
    #:
    #: Supported types:
    #:
    #: - Boolean
    #:
    #: This widget has no custom properties.
    kToggleButton = "Toggle Button"

    #: Displays a boolean with a fixed-size toggle switch.
    #:
    #: Supported types:
    #:
    #: - Boolean
    #:
    #: This widget has no custom properties.
    kToggleSwitch = "Toggle Switch"

    #: Displays an analog input or a raw number with a number bar.
    #:
    #: Supported types:
    #:
    #: - Number
    #: - :class:`.AnalogInput`
    #:
    #: Custom properties:
    #:
    #: ==================== ====== ============= =================================================================
    #: Name                 Type   Default Value Notes
    #: ==================== ====== ============= =================================================================
    #: Min                  Number 0             The minimum value of the bar
    #: Max                  Number 5             The maximum value of the bar
    #: Center               Number 0             The center ("zero") value of the bar
    #: Orientation          String "HORIZONTAL"  The orientation of the bar. One of ``["HORIZONTAL", "VERTICAL"]``
    #: Number of tick marks Number 5             The number of discrete ticks on the bar
    #: ==================== ====== ============= =================================================================
    kVoltageView = "Voltage View"

    #: Displays a :class:`.PowerDistributionPanel`.
    #:
    #: Supported types:
    #:
    #: - :class:`.PowerDistributionPanel`
    #:
    #: =============================== ======= ============= ======================================================
    #: Name                            Type    Default Value Notes
    #: =============================== ======= ============= ======================================================
    #: Show voltage and current values Boolean True          Whether or not to display the voltage and current draw
    #: =============================== ======= ============= ======================================================
    kPowerDistributionPanel = "PDP"

    #: Displays a :class:`.SendableChooser` with a dropdown combo box with a list of options.
    #:
    #: Supported types:
    #:
    #: - :class:`.SendableChooser`
    #:
    #: This widget has no custom properties.
    kComboBoxChooser = "ComboBox Chooser"

    #: Displays a :class:`.SendableChooser` with a dropdown combo box with a toggle button for each available option.
    #:
    #: Supported types:
    #:
    #: - :class:`.SendableChooser`
    #:
    #: This widget has no custom properties.
    kSplitButtonChooser = "Split Button Chooser"

    #: Displays an :class:`.Encoder` displaying its speed, total travelled distance, and its distance per tick.
    #:
    #: Supported types:
    #:
    #: - :class:`.Encoder`
    #:
    #: This widget has no custom properties.
    kEncoder = "Encoder"

    #: Displays a :class:`.SpeedController`.
    #: The speed controller will be controllable from the dashboard when
    #: test mode is enabled, but will otherwise be view-only.
    #:
    #: Supported types:
    #:
    #: - :class:`.PWMSpeedController`
    #: - :class:`.DMC60`
    #: - :class:`.Jaguar`
    #: - :class:`.PWMTalonSRX`
    #: - :class:`.PWMVictorSPX`
    #: - :class:`.SD540`
    #: - :class:`.Spark`
    #: - :class:`.Talon`
    #: - :class:`.Victor`
    #: - :class:`.VictorSP`
    #: - :class:`.SpeedControllerGroup`
    #:
    #: Custom properties:
    #:
    #: =========== ====== ============= =====================================
    #: Name        Type   Default Value Notes
    #: =========== ====== ============= =====================================
    #: Orientation String "HORIZONTAL"  One of ``["HORIZONTAL", "VERTICAL"]``
    #: =========== ====== ============= =====================================
    kSpeedController = "Speed Controller"

    #: Displays a command with a toggle button.
    #: Pressing the button will start the command, and the
    #: button will automatically release when the command completes.
    #:
    #: Supported types:
    #:
    #: - :class:`wpilib.command.command.Command`
    #: - :class:`wpilib.command.commandgroup.CommandGroup`
    #: - Any custom subclass of ``Command`` or ``CommandGroup``
    #:
    #: This widget has no custom properties.
    kCommand = "Command"

    #: Displays a PID command with a checkbox and an editor for the PIDF constants.
    #: Selecting the checkbox will start the command, and the checkbox will
    #: automatically deselect when the command completes.
    #:
    #: Supported types:
    #:
    #: - :class:`wpilib.command.pidcommand.PIDCommand`
    #: - Any custom subclass of ``PIDCommand``
    #:
    #: This widget has no custom properties.
    kPIDCommand = "PID Command"

    #: Displays a PID controller with an editor for the PIDF constants and
    #: a toggle switch for enabling and disabling the controller.
    #:
    #: Supported types:
    #:
    #: - :class:`.PIDController`
    #:
    #: This widget has no custom properties.
    kPIDController = "PID Controller"

    #: Displays an accelerometer with a number bar displaying the magnitude of the acceleration and
    #: text displaying the exact value.
    #:
    #: Supported types:
    #:
    #: - :class:`.AnalogAccelerometer`
    #:
    #: Custom properties:
    #:
    #: =============== ======= ============= ===================================================
    #: Name            Type    Default Value Notes
    #: =============== ======= ============= ===================================================
    #: Min             Number  -1            The minimum acceleration value to display
    #: Max             Number  1             The maximum acceleration value to display
    #: Show text       Boolean True          Show or hide the acceleration values
    #: Precision       Number  2             How many numbers to display after the decimal point
    #: Show tick marks Boolean False         Show or hide the tick marks on the number bars
    #: =============== ======= ============= ===================================================
    kAccelerometer = "Accelerometer"

    #: Displays a 3-axis accelerometer with a number bar for each axis' accleration.
    #:
    #: Supported types:
    #:
    #: - :class:`.ADXL345_I2C`
    #: - :class:`.ADXL345_SPI`
    #: - :class:`.ADXL362`
    #:
    #: Custom properties:
    #:
    #: =============== ============================= ============= ===================================================
    #: Name            Type                          Default Value Notes
    #: =============== ============================= ============= ===================================================
    #: Range           :class:`.Accelerometer.Range` k16G          The accelerometer range
    #: Show value      Boolean                       True          Show or hide the acceleration values
    #: Precision       Number                        2             How many numbers to display after the decimal point
    #: Show tick marks Boolean                       False         Show or hide the tick marks on the number bars
    #: =============== ============================= ============= ===================================================
    k3AxisAccelerometer = "3-Axis Accelerometer"

    #: Displays a gyro with a dial from 0 to 360 degrees.
    #:
    #: Supported types:
    #:
    #: - :class:`.ADXRS450_Gyro`
    #: - :class:`.AnalogGyro`
    #: - Any custom subclass of ``GyroBase`` (such as a MXP gyro)
    #:
    #: Custom properties:
    #:
    #: =================== ======= ============= =============================================
    #: Name                Type    Default Value Notes
    #: =================== ======= ============= =============================================
    #: Major tick spacing  Number  45            Degrees
    #: Starting angle      Number  180           How far to rotate the entire dial, in degrees
    #: Show tick mark ring Boolean True
    #: =================== ======= ============= =============================================
    kGyro = "Gyro"

    #: Displays a relay with toggle buttons for each supported mode (off, on, forward, reverse).
    #:
    #: Supported types:
    #:
    #: - :class:`.Relay`
    #:
    #: This widget has no custom properties.
    kRelay = "Relay"

    #: Displays a differential drive with a widget that displays the speed of each side
    #: of the drivebase and a vector for the direction and rotation of the drivebase.
    #: The widget will be controllable if the robot is in test mode.
    #:
    #: Supported types:
    #:
    #: - :class:`wpilib.drive.differentialdrive.DifferentialDrive`
    #:
    #: Custom properties:
    #:
    #: ===================== ======= ============= ===============================
    #: Name                  Type    Default Value Notes
    #: ===================== ======= ============= ===============================
    #: Number of wheels      Number  4             Must be a positive even integer
    #: Wheel diameter        Number  80            Pixels
    #: Show velocity vectors Boolean True
    #: ===================== ======= ============= ===============================
    kDifferentialDrive = "Differential Drivebase"

    #: Displays a mecanum drive with a widget that displays the speed of each wheel,
    #: and vectors for the direction and rotation of the drivebase.
    #: The widget will be controllable if the robot is in test mode.
    #:
    #: Supported types:
    #:
    #: - :class:`wpilib.drive.mecanumdrive.MecanumDrive`
    #:
    #: Custom properties:
    #:
    #: ===================== ======= ============= =====
    #: Name                  Type    Default Value Notes
    #: ===================== ======= ============= =====
    #: Show velocity vectors Boolean True
    #: ===================== ======= ============= =====
    kMecanumDrive = "Mecanum Drivebase"

    #: Displays a camera stream.
    #:
    #: Supported types:
    #:
    #: - TODO
    #:
    #: Custom properties:
    #:
    #: =============== ======= ============= =====================================================================================
    #: Name            Type    Default Value Notes
    #: =============== ======= ============= =====================================================================================
    #: Show crosshair  Boolean True          Show or hide a crosshair on the image
    #: Crosshair color Color   "white"       Can be a string or a rgba integer
    #: Show controls   Boolean True          Show or hide the stream controls
    #: Rotation        String  "NONE"        Rotates the displayed image. One of ``["NONE", "QUARTER_CW", "QUARTER_CCW", "HALF"]``
    #: =============== ======= ============= =====================================================================================
    kCameraStream = "Camera Stream"



"""
    This is the core of WPILib.
"""

from .accumulatorresult import AccumulatorResult
from .adxl345_i2c import ADXL345_I2C
from .adxl345_spi import ADXL345_SPI
from .adxl362 import ADXL362
from .adxrs450_gyro import ADXRS450_Gyro
from .analogaccelerometer import AnalogAccelerometer
from .analoginput import AnalogInput
from .analoggyro import AnalogGyro
from .analogoutput import AnalogOutput
from .analogpotentiometer import AnalogPotentiometer
from .analogtrigger import AnalogTrigger
from .analogtriggeroutput import AnalogTriggerOutput
from .builtinaccelerometer import BuiltInAccelerometer
from .cameraserver import CameraServer
from .compressor import Compressor
from .controllerpower import ControllerPower
from .counter import Counter
from .digitalglitchfilter import DigitalGlitchFilter
from .digitalinput import DigitalInput
from .digitaloutput import DigitalOutput
from .digitalsource import DigitalSource
from .dmc60 import DMC60
from .doublesolenoid import DoubleSolenoid
from .driverstation import DriverStation
from .encoder import Encoder
from .filter import Filter
from .geartooth import GearTooth
from .gyrobase import GyroBase
from .i2c import I2C
from .interruptablesensorbase import InterruptableSensorBase
from .iterativerobot import IterativeRobot
from .iterativerobotbase import IterativeRobotBase
from .jaguar import Jaguar
from .joystick import Joystick
from .lineardigitalfilter import LinearDigitalFilter
from .livewindow import LiveWindow
from .livewindowsendable import LiveWindowSendable
from .motorsafety import MotorSafety
from .nidecbrushless import NidecBrushless
from .notifier import Notifier
from .pidcontroller import PIDController
from .powerdistributionpanel import PowerDistributionPanel
from .preferences import Preferences
from .pwm import PWM
from .pwmspeedcontroller import PWMSpeedController
from .pwmtalonsrx import PWMTalonSRX
from .pwmvictorspx import PWMVictorSPX
from .relay import Relay
from .resource import Resource
from .robotbase import RobotBase
from .robotcontroller import RobotController
from .robotdrive import RobotDrive
from .robotstate import RobotState
from .samplerobot import SampleRobot
from .sd540 import SD540
from .sendable import Sendable
from .sendablebase import SendableBase
from .sendablebuilder import SendableBuilder
from .sendablechooser import SendableChooser
from .sensorutil import SensorUtil
from .serialport import SerialPort
from .servo import Servo
from .smartdashboard import SmartDashboard
from .solenoidbase import SolenoidBase
from .solenoid import Solenoid
from .spark import Spark
from .speedcontrollergroup import SpeedControllerGroup
from .spi import SPI
from .talon import Talon
from .timedrobot import TimedRobot
from .timer import Timer
from .ultrasonic import Ultrasonic
from .utility import Utility
from .victor import Victor
from .victorsp import VictorSP
from .watchdog import Watchdog
from .xboxcontroller import XboxController

from ._impl.main import run

try:
    from .version import __version__
except ImportError:
    __version__ = "master"



# validated: 2018-09-09 EN 0614913f1abb edu/wpi/first/wpilibj/PWM.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import enum
from typing import Tuple

import hal
import weakref

from .motorsafety import MotorSafety
from .sendablebase import SendableBase
from .resource import Resource
from .sensorutil import SensorUtil
from .sendablebuilder import SendableBuilder

__all__ = ["PWM"]


def _freePWM(handle: hal.DigitalHandle) -> None:
    hal.setPWMDisabled(handle)
    hal.freePWMPort(handle)


class PWM(MotorSafety, SendableBase):
    """Raw interface to PWM generation in the FPGA.
    
    The values supplied as arguments for PWM outputs range from -1.0 to 1.0. They
    are mapped to the hardware dependent values, in this case 0-2000 for the
    FPGA.  Changes are immediately sent to the FPGA, and the update occurs at
    the next FPGA cycle (5.005 ms). There is no delay.

    As of revision 0.1.10 of the FPGA, the FPGA interprets the 0-2000 values as
    follows:
    
    - 2000 = full "forward"
    - 1999 to 1001 = linear scaling from "full forward" to "center"
    - 1000 = center value
    - 999 to 2 = linear scaling from "center" to "full reverse"
    - 1 = minimum pulse width (currently .5ms)
    - 0 = disabled (i.e. PWM output is held low)
    
    kDefaultPwmPeriod is the 1x period (5.05 ms).  In hardware, the period
    scaling is implemented as an output squelch to get longer periods for old
    devices.

    - 20ms periods (50 Hz) are the "safest" setting in that this works for all
      devices
    - 20ms periods seem to be desirable for Vex Motors
    - 20ms periods are the specified period for HS-322HD servos, but work
      reliably down to 10.0 ms; starting at about 8.5ms, the servo sometimes
      hums and get hot; by 5.0ms the hum is nearly continuous
    - 10ms periods work well for Victor 884
    - 5ms periods allows higher update rates for Luminary Micro Jaguar speed
      controllers. Due to the shipping firmware on the Jaguar, we can't run the
      update period less than 5.05 ms.
    """

    class PeriodMultiplier(enum.IntEnum):
        """Represents the amount to multiply the minimum servo-pulse pwm
        period by.
        """

        #: Period Multiplier: don't skip pulses. PWM pulses occur every 5.005 ms
        k1X = 1

        #: Period Multiplier: skip every other pulse. PWM pulses occur every 10.010 ms
        k2X = 2

        #: Period Multiplier: skip three out of four pulses. PWM pulses occur every 20.020 ms
        k4X = 4

    def __init__(self, channel: int) -> None:
        """Allocate a PWM given a channel.

        :param channel: The PWM channel number. 0-9 are on-board, 10-19 are on the MXP port
        """
        super().__init__()
        SendableBase.__init__(self)

        SensorUtil.checkPWMChannel(channel)
        self.channel = channel

        self._handle = hal.initializePWMPort(hal.getPort(channel))
        self.__finalizer = weakref.finalize(self, _freePWM, self._handle)

        self.setDisabled()

        hal.setPWMEliminateDeadband(self.handle, False)

        hal.report(hal.UsageReporting.kResourceType_PWM, channel)
        self.setName("PWM", channel)

        self.setSafetyEnabled(False)

        # Python-specific: Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

    @property
    def handle(self) -> hal.DigitalHandle:
        if not self.__finalizer.alive:
            raise ValueError("Cannot use channel after free() has been called")
        return self._handle

    def close(self) -> None:
        """Free the PWM channel.

        Free the resource associated with the PWM channel and set the value
        to 0.
        """
        super().close()
        if self._handle is None:
            return
        self.__finalizer()
        self._handle = None

    def stopMotor(self):
        self.setDisabled()

    def getDescription(self):
        return "PWM %d" % self.getChannel()

    def enableDeadbandElimination(self, eliminateDeadband: bool) -> None:
        """Optionally eliminate the deadband from a speed controller.

        :param eliminateDeadband: If True, set the motor curve on the Jaguar
            to eliminate the deadband in the middle of the range. Otherwise, keep
            the full range without modifying any values.
        """
        hal.setPWMEliminateDeadband(self.handle, eliminateDeadband)

    def setBounds(
        self,
        max: float,
        deadbandMax: float,
        center: float,
        deadbandMin: float,
        min: float,
    ) -> None:
        """Set the bounds on the PWM pulse widths.

        This sets the bounds on the PWM values for a particular type of
        controller.  The values determine the upper and lower speeds as well
        as the deadband bracket.

        :param max: The max PWM pulse width in ms
        :param deadbandMax: The high end of the deadband range pulse width in ms
        :param center: The center (off) pulse width in ms
        :param deadbandMin: The low end of the deadband pulse width in ms
        :param min: The minimum pulse width in ms
        """
        hal.setPWMConfig(self.handle, max, deadbandMax, center, deadbandMin, min)

    def getRawBounds(self) -> Tuple[int, int, int, int, int]:
        """Gets the bounds on the PWM pulse widths. This Gets the bounds on the PWM values for a
        particular type of controller. The values determine the upper and lower speeds as well
        as the deadband bracket.
        
        :returns: tuple of (max, deadbandMax, center, deadbandMin, min)
        """
        return hal.getPWMConfigRaw(self.handle)

    def getChannel(self) -> int:
        """Gets the channel number associated with the PWM Object.

        :returns: The channel number.
        """
        return self.channel

    def setPosition(self, pos: float) -> None:
        """Set the PWM value based on a position.

        This is intended to be used by servos.

        .. note::

            :func:`setBounds` must be called first.

        :param pos: The position to set the servo between 0.0 and 1.0.
        """
        hal.setPWMPosition(self.handle, pos)

    def getPosition(self) -> float:
        """Get the PWM value in terms of a position.

        This is intended to be used by servos.

        .. note::

            :func:`setBounds` must be called first.

        :returns: The position the servo is set to between 0.0 and 1.0.
        """
        return hal.getPWMPosition(self.handle)

    def setSpeed(self, speed: float) -> None:
        """Set the PWM value based on a speed.

        This is intended to be used by speed controllers.

        .. note::

            :func:`setBounds` must be called first.

        :param speed: The speed to set the speed controller between -1.0 and
            1.0.
        """
        hal.setPWMSpeed(self.handle, speed)

    def getSpeed(self) -> float:
        """Get the PWM value in terms of speed.

        This is intended to be used by speed controllers.

        .. note::

            :func:`setBounds` must be called first.

        :returns: The most recently set speed between -1.0 and 1.0.
        """
        return hal.getPWMSpeed(self.handle)

    def setRaw(self, value: int) -> None:
        """Set the PWM value directly to the hardware.

        Write a raw value to a PWM channel.

        :param value: Raw PWM value.  Range 0 - 255.
        """
        hal.setPWMRaw(self.handle, value)

    def getRaw(self) -> int:
        """Get the PWM value directly from the hardware.

        Read a raw value from a PWM channel.

        :returns: Raw PWM control value.  Range: 0 - 255.
        """
        return hal.getPWMRaw(self.handle)

    def setDisabled(self) -> None:
        """Temporarily disables the PWM output. The next set call will reenable
        the output.
        """
        hal.setPWMDisabled(self.handle)

    def setPeriodMultiplier(self, mult: PeriodMultiplier) -> None:
        """Slow down the PWM signal for old devices.

        :param mult: The period multiplier to apply to this channel
        """
        if mult == PWM.PeriodMultiplier.k4X:
            # Squelch 3 out of 4 outputs
            hal.setPWMPeriodScale(self.handle, 3)
        elif mult == PWM.PeriodMultiplier.k2X:
            # Squelch 1 out of 2 outputs
            hal.setPWMPeriodScale(self.handle, 1)
        elif mult == PWM.PeriodMultiplier.k1X:
            # Don't squelch any outputs
            hal.setPWMPeriodScale(self.handle, 0)
        else:
            raise ValueError("Invalid mult argument '%s'" % mult)

    def setZeroLatch(self) -> None:
        hal.latchPWMZero(self.handle)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("PWM")
        builder.setSafeState(self.setDisabled)
        builder.addDoubleProperty(
            "Value", self.getRaw, lambda value: self.setRaw(int(value))
        )
